from asyncio import TimeoutError as AsyncTimeoutError
from asyncio import create_subprocess_exec, create_subprocess_shell, sleep, wait_for
from typing import Any, Dict, NamedTuple, Optional

_flag_subprocess_running: Dict[str, bool] = {}

DEFAULT_WAIT_FLAG_SLEEP: float = 0.1


async def _wait_flagged_subprocess_running(wait_flag: str) -> None:
    if not _flag_subprocess_running[wait_flag]:
        return
    await sleep(DEFAULT_WAIT_FLAG_SLEEP)


class SubprocessResult(NamedTuple):
    return_code: int
    stdout: Optional[str] = None
    stderr: Optional[str] = None


def _create_subprocess(  # type: ignore
    *args: Any,
    shell: bool,
    stdout: Optional[int],
    stderr: Optional[int],
    limit: int,
    **kwds: Any,
):
    return (
        create_subprocess_shell(*args, stdout=stdout, stderr=stderr, limit=limit, **kwds)  # type: ignore
        if shell
        else create_subprocess_exec(*args, stdout=stdout, stderr=stderr, limit=limit, **kwds)
    )


async def run_subprocess(
    *args: str,
    shell: bool = False,
    encoding: str = 'utf8',
    timeout: Optional[float] = None,
    wait_flag: Optional[str] = None,
    wait_flag_timeout: Optional[float] = None,
    stdout: Optional[int] = -1,  # see asyncio.subprocess.PIPE,
    stderr: Optional[int] = -1,  # see asyncio.subprocess.PIPE,
    limit: int = 2**64,  # see streams._DEFAULT_LIMIT
    **kwds: Dict[str, Any],
) -> SubprocessResult:
    """
    Creates and runs a subprocess with or without shell.

    Provides to time out the Python managed subprocess using asyncio.wait_for.
    Provides to flag Python managed subprocess with timeout as well using unique keys and asyncio.wait_for.

    stdin not supported!

    When shell=True *args will be the "cmd" arg for asyncio.subprocess.create_subprocess_shell
    When shell=False *args will be the "*args" arg (not "program" arg) for asyncio.subprocess.create_subprocess_exec

    Return (return_code, stdout, stderr)
    """
    _ = [kwds.pop(key, None) for key in ['shell', 'encoding', 'timeout', 'wait_flag', 'wait_flag_timeout']]
    if wait_flag:
        if wait_flag not in _flag_subprocess_running:
            _flag_subprocess_running[wait_flag] = True
        elif _flag_subprocess_running.get(wait_flag, True):
            await wait_for(fut=_wait_flagged_subprocess_running(wait_flag=wait_flag), timeout=wait_flag_timeout)
            _flag_subprocess_running[wait_flag] = True
    proc = await _create_subprocess(
        *args,
        shell=shell,  # nosec
        stdout=stdout,
        stderr=stderr,
        limit=limit,
        **kwds,
    )

    try:
        return_code = await wait_for(fut=proc.wait(), timeout=timeout)
    except AsyncTimeoutError:
        proc.terminate()
        return_code = await proc.wait()
    finally:
        if wait_flag and wait_flag in _flag_subprocess_running:
            del _flag_subprocess_running[wait_flag]

    stdout_: Optional[bytes] = await proc.stdout.read()
    stderr_: Optional[bytes] = await proc.stderr.read()

    return SubprocessResult(
        return_code=return_code,
        stdout=stdout_.strip().decode(encoding) if stdout_ else None,
        stderr=stderr_.strip().decode(encoding) if stderr_ else None,
    )



from logging import NOTSET, Formatter, Logger, StreamHandler, getLogger
from os import getenv
from typing import Any, Dict, List, Optional, Type, TypeVar, Union, cast


def get_env(key: str, default: Optional[str] = None, cast_default_to_str: bool = True) -> Optional[str]:
    """Get an environment variable, return default if it is empty or doesn't exist."""
    value = getenv(key, default)
    if value is None or len(value) == 0:
        value = str(default) if cast_default_to_str else default
    return value


def get_str_env(key: str, default: str = '') -> str:
    return cast(str, get_env(key=key, default=default, cast_default_to_str=True))


_boolean_positive_values: List[str] = ['True', 'true', 'yes', 'Y', 'y', '1']


def get_bool_env(key: str, default: Union[bool, int] = False) -> bool:
    return get_env(key=key, default=str(int(default)), cast_default_to_str=False) in _boolean_positive_values


def get_int_env(key: str, default: Union[bool, int] = 0) -> int:
    val = cast(str, get_env(key=key, default=str(int(default)), cast_default_to_str=False))
    return int(val) if val.isdigit() else default


def get_float_env(key: str, default: Union[bool, int, float] = 0) -> float:
    val = cast(str, get_env(key=key, default=str(float(default)), cast_default_to_str=False))
    return float(val) if val.isdigit() or val.replace('.', '').isdigit() else default


def get_list_str_env(
    key: str,
    default: Optional[List[str]] = None,
    *,
    delimiter: str = ',',
    allow_empty: bool = True,
) -> List[str]:
    val = cast(str, get_env(key=key, default='' if allow_empty else delimiter.join(default or [])))
    return [] if allow_empty and (val is None or len(val) == 0) else val.split(delimiter)


def get_simple_logger(
    name: Optional[str] = None,
    level: Union[str, int] = NOTSET,
    fmt: str = '[%(asctime)s] - %(name)s - %(levelname)s - %(message)s',
) -> Logger:  # pragma: no cover
    logger = getLogger(name)
    logger.setLevel(level)
    handler = StreamHandler()
    handler.setLevel(level)
    formatter = Formatter(fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


_T = TypeVar('_T')
_ENV: Optional[Dict[str, Any]] = None


def env(key: Optional[str] = None, typ: Optional[Type[_T]] = None) -> _T:
    """Get full environment variables resolved if no argument given or env var matching key and optional type given."""
    global _ENV
    if _ENV is None:
        _ENV = env.resolver()  # type: ignore
    if key is None:
        return _ENV  # type: ignore
    if not isinstance(_ENV, dict):
        raise ValueError('"_ENV" variable must be a dict. Check env.resolver method.')
    if key not in _ENV:
        raise KeyError(
            '<{0}{1}> does not exist as environment variable'.format(key, ': {0}'.format(typ.__name__) if typ else '')
        )
    val = _ENV[key]
    if typ and not isinstance(val, (typ,)):
        raise TypeError(
            '<{0}{1}> does not exist as environment variable'.format(key, ': {0}'.format(typ.__name__) if typ else '')
        )
    return val  # type: ignore


env.resolver = lambda: None  # type: ignore



from abc import ABC, abstractmethod
from calendar import timegm
from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Any, Dict, List, Type, Union
from uuid import uuid4

from .errors import EventMapperNotFoundError


@dataclass
class Event:
    """Class for keeping track of an Event."""

    @dataclass
    class Attributes:
        """Class for keeping track of an Attributes of Event."""

    @dataclass
    class Meta:
        """Class for keeping track of a Metas of Event."""

        __slots__ = ('id', 'type', 'occurred_on')
        id: str
        type: str
        occurred_on: int

    attributes: Attributes = field(default_factory=lambda: Event.Attributes())
    meta: Meta = field(
        default_factory=lambda: Event.Meta(
            id=str(uuid4()),
            type='event',
            occurred_on=timegm(datetime.utcnow().utctimetuple()),
        )
    )


class EventMapper:
    __slots__ = ('event_type', 'service_name', 'event_name')

    event_type: Type[Event]
    service_name: str
    event_name: str

    def belongs_to(self, msg: Event) -> bool:
        return isinstance(msg, self.event_type)

    def encode(self, msg: Event) -> Dict[str, Any]:
        return {
            **asdict(msg.meta),
            'attributes': self.map_attributes(msg.attributes),
            'meta': {'message': f'{self.service_name}.{self.event_name}'},
        }

    def decode(self, data: Dict[str, Any]) -> Event:
        attributes = self.event_type.Attributes(**data['attributes'])
        meta = self.event_type.Meta(id=data['id'], type=data['type'], occurred_on=data['occurred_on'])
        return self.event_type(attributes=attributes, meta=meta)

    @staticmethod
    def map_attributes(attributes: Event.Attributes) -> Dict[str, Any]:
        return asdict(attributes)


def find_event_mapper_by_name(name: str, mappers: List[EventMapper]) -> EventMapper:
    for mapper in mappers:
        if f'{mapper.service_name}.{mapper.event_name}' == name:
            return mapper
    raise EventMapperNotFoundError.create(detail={'name': name})


def find_event_mapper_by_type(msg: Event, mappers: List[EventMapper]) -> EventMapper:
    for mapper in mappers:
        if mapper.belongs_to(msg):
            return mapper
    raise EventMapperNotFoundError.create(detail={'type': str(type(msg))})


class ConfigEventMappers:
    _mappers: List[EventMapper]

    def __init__(self, mappers: List[EventMapper]) -> None:
        self._mappers = mappers

    def add(self, mappers: Union[EventMapper, List[EventMapper]]) -> None:
        if isinstance(mappers, EventMapper):
            self._mappers.append(mappers)
        elif isinstance(mappers, list):
            for mapper in mappers:
                self._mappers.append(mapper)

    def all(self) -> List[EventMapper]:
        return self._mappers


class EventPublisher(ABC):
    @abstractmethod
    async def publish(self, events: List[Event]) -> None:
        pass  # pragma: no cover


class EventPublishers(EventPublisher):
    _publishers: List[EventPublisher]

    def __init__(self, publishers: List[EventPublisher]) -> None:
        self._publishers = publishers

    def add(self, publishers: Union[EventPublisher, List[EventPublisher]]) -> None:
        if not isinstance(publishers, list):
            publishers = [publishers]
        for publisher in publishers:
            self._publishers.append(publisher)

    async def publish(self, events: List[Event]) -> None:
        for publisher in self._publishers:
            await publisher.publish(events)


class EventHandler(ABC):
    @abstractmethod
    def subscribed_to(self) -> List[Type[Event]]:
        pass  # pragma: no cover

    @abstractmethod
    async def handle(self, events: List[Event]) -> None:
        pass  # pragma: no cover


class EventBus(ABC):
    @abstractmethod
    async def notify(self, events: List[Event]) -> None:
        pass  # pragma: no cover


class SimpleEventBus(EventBus):
    _handlers: List[EventHandler]

    def __init__(self, handlers: List[EventHandler]):
        self._handlers = handlers

    def add_handler(self, handler: Union[EventHandler, List[EventHandler]]) -> None:
        if not isinstance(handler, list):
            handler = [handler]
        for handler_ in handler:
            self._handlers.append(handler_)

    async def notify(self, events: List[Event]) -> None:
        for event in events:
            for handler in self._handlers:
                for event_type in handler.subscribed_to():
                    if isinstance(event, event_type):
                        await handler.handle([event])


class InternalEventPublisher(EventPublisher):
    __slots__ = '_event_bus'

    def __init__(self, event_bus: EventBus) -> None:
        self._event_bus = event_bus

    async def publish(self, events: List[Event]) -> None:
        await self._event_bus.notify(events=events)



from abc import ABC
from typing import List

from .events import Event


class Aggregate(ABC):
    pass


class AggregateRoot(Aggregate):
    _events: List[Event]

    def __init__(self) -> None:
        self._events = []

    def pull_aggregate_events(self) -> List[Event]:
        _events = self._events
        self._events = []
        return _events

    def record_aggregate_event(self, event: Event) -> None:
        self._events.append(event)



from json import dumps
from typing import Any, Dict, Optional
from uuid import uuid4


def raise_(err: BaseException) -> BaseException:
    raise err  # pragma: no cover


class BaseError(Exception):
    _id: str
    _code: str = 'code'
    _title: str = 'title'
    _detail: str
    _meta: Dict[str, Any]

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args)
        self._id = kwargs.get('id', str(uuid4()))
        self._code = kwargs.get('code', self._code)
        self._title = kwargs.get('title', self._title)
        self._detail = dumps(kwargs.get('detail', {}))
        self._meta = kwargs.get('meta', {})
        self.ensure_there_is_not_a_system_exit()

    @classmethod
    def create(
        cls,
        detail: Optional[Dict[str, Any]] = None,
        meta: Optional[Dict[str, Any]] = None,
        **kwargs: Dict[str, Any],
    ) -> 'BaseError':
        return cls(detail=detail or {}, meta=meta or {}, **kwargs)

    def with_exception(self, err: BaseException) -> 'BaseError':
        self._meta.update({'exception': str(err), 'exception_type': str(type(err))})
        self.ensure_there_is_not_a_system_exit()
        return self

    def ensure_there_is_not_a_system_exit(self) -> None:
        if self._meta.get('exception_type', None) == '<class \'SystemExit\'>':
            raise SystemExit(self._meta.get('exception'))

    def id(self) -> str:
        return self._id

    def code(self) -> str:
        return self._code

    def title(self) -> str:
        return self._title

    def detail(self) -> str:
        return self._detail

    def meta(self) -> Dict[str, Any]:
        return self._meta

    def __str__(self) -> str:
        return dumps(
            {
                'id': self._id,
                'code': self._code,
                'title': self._title,
                'detail': self._detail,
                'meta': self._meta,
            },
            indent=2,
        )


class NotFoundError(BaseError):
    _code = 'not_found'
    _title = 'Not found'


class ConflictError(BaseError):
    _code = 'conflict'
    _title = 'Conflict'


class BadRequestError(BaseError):
    _code = 'bad_request'
    _title = 'Bad Request'


class UnauthorizedError(BaseError):
    _code = 'unauthorized'
    _title = 'Unauthorized'


class ForbiddenError(BaseError):
    _code = 'forbidden'
    _title = 'Forbidden'


class UnknownError(BaseError):
    _code = 'unknown'
    _title = 'Unknown error'


class IdInvalidError(ConflictError):
    _code = 'id_invalid'
    _title = 'Invalid id'


class TimestampInvalidError(ConflictError):
    _code = 'timestamp_invalid'
    _title = 'Invalid timestamp'


class DateTimeInvalidError(ConflictError):
    _code = 'datetime_invalid'
    _title = 'Invalid datetime'


class EventMapperNotFoundError(NotFoundError):
    _code = 'event_mapper_not_found'
    _title = 'Event Mapper not found'


class EventNotPublishedError(ConflictError):
    _code = 'event_not_published'
    _title = 'Event not published'


class CommandNotRegisteredError(NotFoundError):
    _code = 'command_not_registered_error'
    _title = 'Command not registered'


class QueryNotRegisteredError(NotFoundError):
    _code = 'query_not_registered_error'
    _title = 'Query not registered'



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Type, Union

from .errors import CommandNotRegisteredError, QueryNotRegisteredError


class Command(ABC):
    pass


class CommandHandler(ABC):
    @abstractmethod
    def subscribed_to(self) -> Type[Command]:
        pass  # pragma: no cover

    @abstractmethod
    async def handle(self, command: Command) -> None:
        pass  # pragma: no cover


class CommandBus(ABC):
    @abstractmethod
    async def dispatch(self, command: Command) -> None:
        pass  # pragma: no cover


class SimpleCommandBus(CommandBus):
    _handlers: List[CommandHandler]

    def __init__(self, handlers: List[CommandHandler]) -> None:
        self._handlers = handlers

    def add_handler(self, handler: Union[CommandHandler, List[CommandHandler]]) -> None:
        if not isinstance(handler, list):
            handler = [handler]
        for handler_ in handler:
            self._handlers.append(handler_)

    async def dispatch(self, command: Command) -> None:
        handlers = [handler for handler in self._handlers if isinstance(command, handler.subscribed_to())]
        if len(handlers) != 1:
            raise CommandNotRegisteredError.create(detail={'command': command.__class__.__name__})
        await handlers[0].handle(command)


class Query(ABC):
    pass


class Response(Dict[str, Any]):
    pass


OptionalResponse = Optional[Union[Any, Response, List[Response]]]


class QueryHandler(ABC):
    @abstractmethod
    def subscribed_to(self) -> Type[Query]:
        pass  # pragma: no cover

    @abstractmethod
    async def handle(self, query: Query) -> OptionalResponse:
        pass  # pragma: no cover


class QueryBus(ABC):
    @abstractmethod
    async def ask(self, query: Query) -> OptionalResponse:
        pass  # pragma: no cover


class SimpleQueryBus(QueryBus):
    _handlers: List[QueryHandler]

    def __init__(self, handlers: List[QueryHandler]) -> None:
        self._handlers = handlers

    def add_handler(self, handler: Union[QueryHandler, List[QueryHandler]]) -> None:
        if not isinstance(handler, list):
            handler = [handler]
        for handler_ in handler:
            self._handlers.append(handler_)

    async def ask(self, query: Query) -> OptionalResponse:
        handlers = [handler for handler in self._handlers if isinstance(query, handler.subscribed_to())]
        if len(handlers) != 1:
            raise QueryNotRegisteredError.create(detail={'query': query.__class__.__name__})
        return await handlers[0].handle(query)



from typing import Any, Dict, List, Optional, Type, Union
from unittest.mock import AsyncMock, MagicMock, Mock, patch


SanitizeObject = Union[Dict[Any, Any], List[Any]]


def sanitize_objects(source: SanitizeObject, affected: SanitizeObject) -> SanitizeObject:  # pragma: no cover
    for key, value in list(affected.items()) if isinstance(affected, dict) else enumerate(affected):
        if not isinstance(affected, list) and key not in source:
            affected.pop(key)
        elif isinstance(value, (dict, list)):
            affected[key] = sanitize_objects(source[key], value)
    return affected


def mock(
    target: Union[str, object],
    attributes: Optional[List[str]] = None,
    target_mock_type: Type[Mock] = AsyncMock,
    attribute_mock_type: Type[Mock] = AsyncMock,
) -> Mock:
    target_async_mock = target_mock_type()
    if not attributes:
        patch(
            target=f'{target.__module__}.{target.__name__}' if isinstance(target, object) else target,  # type: ignore
            side_effect=target_async_mock,
        )
        return target_async_mock
    for attribute in attributes:
        attribute_async_mock = attribute_mock_type()
        patch.object(
            target=target,
            attribute=attribute,
            side_effect=attribute_async_mock,
        )
        target_async_mock[attribute] = attribute_async_mock
    return target_async_mock


async_mock = mock


def sync_mock(
    target: Union[str, object],
    attributes: Optional[List[str]] = None,
    target_mock_type: Type[Mock] = MagicMock,
    attribute_mock_type: Type[Mock] = MagicMock,
) -> Mock:
    return mock(
        target=target,
        attributes=attributes,
        target_mock_type=target_mock_type,
        attribute_mock_type=attribute_mock_type,
    )



# type: ignore
# pylint: skip-file
from .aggregates import Aggregate, AggregateRoot
from .cqrs import (
    Command,
    CommandBus,
    CommandHandler,
    OptionalResponse,
    Query,
    QueryBus,
    QueryHandler,
    Response,
    SimpleCommandBus,
    SimpleQueryBus,
)
from .errors import (
    BadRequestError,
    BaseError,
    CommandNotRegisteredError,
    ConflictError,
    DateTimeInvalidError,
    EventMapperNotFoundError,
    EventNotPublishedError,
    ForbiddenError,
    IdInvalidError,
    NotFoundError,
    QueryNotRegisteredError,
    TimestampInvalidError,
    UnauthorizedError,
    UnknownError,
)
from .events import (
    ConfigEventMappers,
    Event,
    EventBus,
    EventHandler,
    EventMapper,
    EventMapperNotFoundError,
    EventPublisher,
    EventPublishers,
    InternalEventPublisher,
    SimpleEventBus,
    find_event_mapper_by_name,
    find_event_mapper_by_type,
)
from .subprocess import SubprocessResult, run_subprocess  # nosec
from .utils import (
    env,
    get_bool_env,
    get_env,
    get_float_env,
    get_int_env,
    get_list_str_env,
    get_simple_logger,
    get_str_env,
)
from .value_objects import Id, StrDateTime, Timestamp

__version__ = '1.4.0'

__all__ = (
    # aggregates
    'Aggregate',
    'AggregateRoot',
    # cqrs
    'Command',
    'CommandHandler',
    'CommandBus',
    'SimpleCommandBus',
    'Query',
    'Response',
    'OptionalResponse',
    'QueryHandler',
    'QueryBus',
    'SimpleQueryBus',
    # errors
    'BaseError',
    'NotFoundError',
    'ConflictError',
    'BadRequestError',
    'UnauthorizedError',
    'ForbiddenError',
    'UnknownError',
    'IdInvalidError',
    'TimestampInvalidError',
    'DateTimeInvalidError',
    'EventNotPublishedError',
    'CommandNotRegisteredError',
    'QueryNotRegisteredError',
    # events
    'Event',
    'EventMapper',
    'EventPublisher',
    'EventHandler',
    'EventBus',
    'SimpleEventBus',
    'find_event_mapper_by_name',
    'find_event_mapper_by_type',
    'EventPublishers',
    'ConfigEventMappers',
    'EventMapperNotFoundError',
    'InternalEventPublisher',
    # utils
    'get_env',
    'get_str_env',
    'get_bool_env',
    'get_int_env',
    'get_float_env',
    'get_list_str_env',
    'get_simple_logger',
    'env',
    # value_objects
    'Id',
    'Timestamp',
    'StrDateTime',
    # subprocess,
    'SubprocessResult',
    'run_subprocess',
)
